"""
Custom fields for storing RDF primitives.

Based on http://blog.elsdoerfer.name/2008/01/08/fuzzydates-or-one-django-model-field-multiple-database-columns/
"""  # noqa: E501
from django.db import models
from rdflib.graph import Graph
from rdflib.term import BNode, URIRef, Literal


class LiteralField(models.TextField):
    """
    Custom field for storing literals.
    """

    description = "Field for storing Literals, including their type and language"  # noqa: 501

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value

        parts = value.split('^^')
        if len(parts) != 3:
            raise ValueError("Wrong value: {}".format(value))
        return Literal(parts[0], parts[1] or None, parts[2] or None)

    def to_python(self, value):
        if isinstance(value, Literal):
            return value

        if value is None:
            return value

        parts = value.split('^^')
        if len(parts) != 3:
            raise ValueError("Wrong value: {0}".format(value))
        return Literal(parts[0], parts[1] or None, parts[2] or None)

    def get_prep_value(self, value):
        if not isinstance(value, Literal):
            raise TypeError("Value {} has the wrong type: {}".format(
                value, value.__class__)
            )

        return (
            str(value) +
            "^^" +
            str(value.language or '') +
            "^^" +
            str(value.datatype or '')
        )


def deserialize_uri(value):
    """
    Deserialize a representation of a BNode or URIRef.
    """
    if isinstance(value, BNode):
        return value
    if isinstance(value, URIRef):
        return value
    if not value:
        return None
    if not isinstance(value, str):
        raise ValueError("Cannot create URI from {} of type {}".format(
            value, value.__class__)
        )
    if value.startswith("_:"):
        return BNode(value[2:])
    return URIRef(value)


def serialize_uri(value):
    """
    Serialize a BNode or URIRef.
    """
    if isinstance(value, BNode):
        return value.n3()
    if isinstance(value, URIRef):
        return str(value)
    # neccessary for migrations in some db backends (e.g. postgres)
    if value is None:
        return None
    raise ValueError("Cannot get prepvalue for {} of type {}".format(
        value, value.__class__)
    )


class URIField(models.CharField):
    """
    Custom field for storing URIRefs and BNodes.

    URIRefs are stored as themselves; BNodes are stored in their Notation3 serialization.
    """  # noqa: E501

    description = "Field for storing URIRefs and BNodes."

    def __init__(self, *args, **kwargs):
        if 'max_length' not in kwargs:
            kwargs['max_length'] = 500
        super(URIField, self).__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection):
        return deserialize_uri(value)

    def to_python(self, value):
        return deserialize_uri(value)

    def get_prep_value(self, value):
        return serialize_uri(value)


class GraphReferenceField(models.CharField):
    """
    Custom field for storing graph references.
    """

    description = "Field for storing references to Graphs"

    def from_db_value(self, value, expression, connection):
        if isinstance(value, Graph):
            return value.identifier

        return deserialize_uri(value)

    def to_python(self, value):
        if isinstance(value, Graph):
            return value.identifier

        return deserialize_uri(value)

    def get_prep_value(self, value):
        if isinstance(value, Graph):
            return serialize_uri(value.identifier)

        return serialize_uri(value)



__all__ = ["RdflibDjangoConfig"]

from django.apps import AppConfig
from django.db.models.signals import post_save

extra_namespaces = [
    {
        "prefix": "xml",
        "fixed": True,
        "uri": "http://www.w3.org/XML/1998/namespace"
    }
]


def UpdateNamespaces(sender, instance, created, **kwargs):
    if not created:
        return
    from .models import NamespaceModel
    from django.db.models import Q
    from rdflib import namespace

    for val in extra_namespaces:
        if not val.get("fixed", False):
            if not NamespaceModel.objects.filter(
                Q(uri=val["uri"]) | Q(prefix=val["prefix"]),
                store=instance
            ):
                NamespaceModel.objects.create(
                    store=instance,
                    uri=val["uri"],
                    prefix=val["prefix"]
                )
        else:
            if not NamespaceModel.objects.filter(
                Q(uri=val["uri"]) & Q(prefix=val["prefix"]),
                store=None
            ):
                # cleanup old namespaces
                NamespaceModel.objects.filter(
                    Q(uri=val["uri"]) | Q(prefix=val["prefix"])
                ).delete()
                NamespaceModel.objects.create(
                    store=None,
                    uri=val["uri"],
                    prefix=val["prefix"]
                )

    for key in namespace.__all__:
        val = getattr(namespace, key)
        if isinstance(val, namespace.Namespace):
            if not NamespaceModel.objects.filter(
                Q(uri=val.uri) | Q(prefix=key.lower()),
                store=instance
            ):
                NamespaceModel.objects.create(
                    prefix=key.lower(),
                    uri=val.uri,
                    store=instance
                )
        elif isinstance(val, namespace.ClosedNamespace):
            if not NamespaceModel.objects.filter(
                Q(uri=val.uri) & Q(prefix=key.lower()),
                store=None
            ):
                NamespaceModel.objects.filter(
                    Q(uri=val.uri) | Q(prefix=key.lower()),
                    store=None
                ).delete()
                NamespaceModel.objects.create(
                    prefix=key.lower(),
                    uri=val.uri,
                    store=None
                )
            # cleanup old namespaces
            NamespaceModel.objects.filter(
                Q(uri=val.uri) | Q(prefix=key.lower())
            ).exclude(store__isnull=True).delete()


class RdflibDjangoConfig(AppConfig):
    name = "rdflib_django"

    def ready(self):
        from .models import Store
        post_save.connect(
            UpdateNamespaces, sender=Store
        )



"""
Settings for testing the application.
"""
from django.conf.global_settings import *  # noqa
import os

DEBUG = True
SECRET_KEY = "FKJSLSOIIDSPOSOPDS"

DJANGO_RDFLIB_DEVELOP = True

DB_PATH = os.path.abspath(os.path.join(__file__, '..', 'rdflib_django.db'))

# TEST_RUNNER = "unittest.TextTestRunner"

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': DB_PATH,
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
        }
}

SITE_ID = 1

STATIC_URL = '/static/'

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    # required for checking bugs in admin.py
    'django.contrib.admin',

    'rdflib_django',
    )
ROOT_URLCONF = 'test.urls'

MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.i18n',
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
        },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
        },
    'loggers': {
        '': {
            'handlers': ['console'],
            'propagate': True,
            'level': 'INFO',
            },
        }
}
